from .helpers import give, iterate, log
from .tools import configurable, instrument_definition
from .version import version as __version__

__all__ = [
    "give",
    "iterate",
    "log",
    "configurable",
    "instrument_definition",
    "__version__",
]



import json
import os
import re
import select
from copy import deepcopy
from pathlib import Path

import pytest

from voir.overseer import Overseer

_progdir = Path(__file__).parent / "programs"


@pytest.fixture
def progdir():
    return _progdir


def _format(x):
    idx = x.get("index", 1)
    title = x.pipe or "---"
    content = None
    if x.event == "line":
        content = x.data
    elif x.event == "binary":
        content = f"{x.data}\n"
    elif x.event == "data":
        content = f"{json.dumps(x.data)}\n"
    elif x.event == "start" or x.event == "end":
        title = f"{x.event}\n"
    else:
        title = f"{title}.{x.event}"
        content = f"{json.dumps(x.data)}\n"

    if content:
        return f"#{idx} {title}: {content}"
    else:
        return f"#{idx} {title}"


run_program_template = """Readable
=========
{readable}
Raw
=========
{raw}
"""


order = ["start", "stdout", "stderr", "data", "end"]


def _order_key(entry):
    return order.index(entry.pipe or entry.event)


@pytest.fixture
def run_program(file_regression):
    from voir.proc import Multiplexer

    def run(
        argv, info={}, voirfile=None, env=None, reorder=True, constructor=None, **kwargs
    ):
        if env is None:
            env = os.environ
        if voirfile is not None:
            env = {**os.environ, "VOIRFILE": voirfile}
        mp = Multiplexer(timeout=None, constructor=constructor)
        mp.start(argv, info=info, cwd=_progdir, env=env, buffered=False, **kwargs)
        results = list(mp)
        if reorder:
            results.sort(key=_order_key)
        for r in results:
            # Patch out the times because they will change from a run to the other
            if r.event in ("start", "end"):
                r.data["time"] = "X"

        readable = "".join(_format(deepcopy(x)) for x in results)
        raw = "\n".join(
            x.json() if x.event != "binary" else str(x.dict()) for x in results
        )
        file_regression.check(run_program_template.format(readable=readable, raw=raw))

    return run


@pytest.fixture
def outlines(capsys):
    def read():
        return [x for x in capsys.readouterr().out.split("\n") if x]

    return read


output_summary_template = """##########
# stdout #
##########
{out}
##########
# stderr #
##########
{err}
##########
#  data  #
##########
{data}
"""


@pytest.fixture
def output_summary(capsys, capdata):
    def calc():
        oe = capsys.readouterr()
        dat = capdata()
        txt = output_summary_template.format(out=oe.out, err=oe.err, data=dat)
        txt = re.sub(string=txt, pattern='File "[^"]*", line [0-9]+', repl="<redacted>")
        txt = re.sub(string=txt, pattern=r"[ \^]+\n", repl="")
        return txt

    return calc


@pytest.fixture
def check_all(output_summary, file_regression):
    yield
    file_regression.check(output_summary())


@pytest.fixture
def data_fds():
    r, w = os.pipe()
    yield r, w


@pytest.fixture
def capdata(data_fds):
    r, w = data_fds
    with open(r, "r") as reader:

        def read():
            r, _, _ = select.select([reader], [], [], 0)
            if reader in r:
                return reader.read()
            else:
                return ""

        yield read


@pytest.fixture
def ov(data_fds):
    r, w = data_fds
    return Overseer(instruments=[], logfile=w)


@pytest.fixture
def ov_nodata():
    return Overseer(instruments=[])



import json
from dataclasses import dataclass

from voir.proc import LogEntry, run
from voir.smuggle import encode_as_escape_sequence


@dataclass
class LogWithIndex(LogEntry):
    index: int = 0


def test_multiplexer(run_program):
    run_program(["python", "datafd.py"], info={"index": 1}, constructor=LogWithIndex)


def test_run():
    results = run(
        ["echo", "hello"], timeout=None, info={"index": 1}, constructor=LogWithIndex
    )
    found = False
    for entry in results:
        assert isinstance(entry, LogWithIndex) and entry.index == 1
        if entry.pipe == "stdout":
            assert entry.data == "hello\n"
            found = True
    assert found


def test_run_stdout():
    secret = {"sekret": True}
    results = run(
        [
            "echo",
            "hello",
            encode_as_escape_sequence(f"{json.dumps(secret)}\n"),
            "world",
        ],
        timeout=None,
        info={"index": 1},
        constructor=LogWithIndex,
        use_stdout=True,
    )
    found = 0
    for entry in results:
        assert isinstance(entry, LogWithIndex) and entry.index == 1
        if entry.pipe == "stdout":
            assert entry.data == "hello  world\n"
            found += 1
        if entry.pipe == "data":
            assert entry.data == secret
            found += 1
    assert found == 2



import time

import pytest

from voir.instruments.gpu import (
    NotAvailable,
    get_backends,
    get_gpu_info,
    select_backend,
)
from voir.instruments.metric import rate

from .common import program


class Collect:
    def __init__(self):
        self.results = []

    def __call__(self, ov):
        yield ov.phases.init

        ov.given.print()
        ov.given["?rate"].map(round) >> self.results.append


@pytest.fixture
def faketime(monkeypatch):
    current_time = [0]

    def sleep(n):
        current_time[0] += n

    def nano():
        return current_time[0] * 1_000_000_000

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(time, "time", lambda: current_time[0])
    monkeypatch.setattr(time, "time_ns", nano)


@pytest.mark.parametrize("interval", [1, 2])
def test_rate(ov, interval, faketime):
    c = Collect()

    ov.require(c)
    ov.require(rate(interval=interval, batch_size_calc=len))

    ov([program("rates")])
    assert c.results == [100] * (10 // interval)


@pytest.mark.parametrize("interval", [1, 2, 5])
def test_sync(ov, interval, faketime):
    def sync():
        time.sleep(0.9)

    c = Collect()

    ov.require(c)
    ov.require(rate(interval=interval, batch_size_calc=len, sync=sync))

    expected_time = 10 * 0.1 + (10 // interval) * 0.9

    ov([program("rates")])
    assert c.results == [round(100 / expected_time)] * (10 // interval)


def test_gpu_backend():
    assert list(sorted(get_backends())) == ["cpu", "cuda", "hpu", "rocm", "xpu"]


def test_gpu_info_cpu():
    assert get_gpu_info() == {"arch": "cpu", "gpus": {}}


def test_select_backend_rocm():
    with pytest.raises(NotAvailable):
        select_backend("rocm")


def test_select_backend_cuda():
    with pytest.raises(NotAvailable):
        select_backend("cuda")



from dataclasses import dataclass, field

import pytest

from voir.argparse_ext import ExtendedArgumentParser


@dataclass
class Muffin:
    __help__ = "Lovely muffinness"

    class RedHerring:
        def hello(x):
            # wow!
            z = x * x
            return z

    moistness: int = 10
    has_chocolate: bool = True


@dataclass
class Configuration:
    # Hello there
    helloes: int = 3

    # Flag this
    flag: bool = True

    flog: bool = False  # Flog this

    # Mystery.
    x: int = 0

    greeting: str = "hello"
    """The greeting to use!"""

    scone: Muffin = field(default_factory=Muffin)


@pytest.fixture
def configuration():
    p = ExtendedArgumentParser()
    p.add_from_model("cfg", Configuration())

    def run(args):
        return p.parse_args(args.split()).cfg

    run.parser = p

    return run


@pytest.fixture
def based_configuration():
    p = ExtendedArgumentParser()
    p.merge_base_config({"cfg": {"helloes": 77, "scone": {"has_chocolate": False}}})
    p.merge_base_config(
        {
            "cfg": {
                "helloes": 88,
                "flag": False,
            }
        }
    )
    p.add_from_model("cfg", Configuration)

    def run(args):
        return p.parse_args(args.split()).cfg

    run.parser = p

    return run


def test_parser(configuration):
    assert configuration("") == Configuration()
    assert configuration("--helloes 5") == Configuration(helloes=5)
    assert configuration("--flag") == Configuration(flag=True)
    assert configuration("--no-flag") == Configuration(flag=False)
    assert configuration("--flog") == Configuration(flog=True)
    assert configuration("--no-flog") == Configuration(flog=False)
    assert configuration("--scone.moistness 33") == Configuration(
        scone=Muffin(moistness=33)
    )
    with pytest.raises(Exception):
        configuration.parser.merge_base_config(Configuration())


def test_parser_with_base(based_configuration):
    assert based_configuration("") == Configuration(
        helloes=88, flag=False, scone=Muffin(has_chocolate=False)
    )
    assert based_configuration("--scone.has-chocolate") == Configuration(
        helloes=88, flag=False, scone=Muffin(has_chocolate=True)
    )


def test_help(configuration, file_regression, capsys):
    with pytest.raises(SystemExit):
        configuration("-h")

    file_regression.check(
        capsys.readouterr().out.replace("options:", "optional arguments:")
    )


def test_help_with_base(based_configuration, file_regression, capsys):
    with pytest.raises(SystemExit):
        based_configuration("-h")

    file_regression.check(
        capsys.readouterr().out.replace("options:", "optional arguments:")
    )



"""Program runner.

Exports :func:`run(argv, ...)<run>` which can be used to run a program and iterate
over its stdout, stderr and data.

Exports the :class:`Multiplexer` class which can run multiple programs in parallel
and yield a unified stream of their stdout, stderr and any data that they generate
through ``voir``.
"""

import json
import os
import select
import subprocess
import time
from dataclasses import dataclass
from typing import Callable

from voir.smuggle import Decoder, MultimodalFile


@dataclass
class _Stream:
    pipe: object
    info: dict
    deserializer: Callable = None


@dataclass
class LogEntry:
    """An entry yielded by iterating over a :class:`Multiplexer`."""

    event: str
    """Name of the event (e.g. 'start', 'line', 'data', etc)."""

    data: object
    """Data associated to the event."""

    pipe: str = None
    """Pipe on which the event was produced (e.g. 'stdout', 'stderr', 'data')."""

    def get(self, item, default):
        return getattr(self, item, default)

    def dict(self):
        """Convert this entry to a plain dictionary."""
        return dict(self.__dict__)

    def json(self):
        """Convert this entry to JSON."""
        return json.dumps(self.__dict__)


def run(argv, info, timeout=None, constructor=None, env=None, **options):
    """Run a program.

    The result is a :class:`Multiplexer` that can be iterated over in order to
    get events corresponding to the start and end of the program as well as its
    stdout/stderr and data generated via ``voir``.

    Arguments:
        argv: The list of arguments.
        info: A dictionary of extra information that will be embedded in the
            ``LogEntry`` objects that are generated. The class given as the ``constructor``
            parameter should be able to take this information as keyword arguments
            in its ``__init__`` function.
        timeout: Timeout to use when using ``select`` to block on the next input.
        constructor: The subtype of :class:`LogEntry` to build log entries with.
            By default it is just ``LogEntry``.
        env: Environment variables to set.
        options: Other options to pass to :meth:`Multiplexer.start`.

    Returns:
        A Multiplexer running that program.
    """
    mp = Multiplexer(timeout=timeout, constructor=constructor)
    mp.start(argv, info=info, env=env, **options)
    return mp


class Multiplexer:
    """Run multiple programs in parallel and yield a unified stream of events.

    Iterating over a Multiplexer generates a sequence of :class:`LogEntry`:

    * The first event has ``event == "start"``.
    * Each line on stdout produces ``event == "line" and pipe == "stdout"``
    * Data logged through Voir produces ``event == "data" and pipe == "data" and data == the_data``.
    * The last event has ``event == "end"``.

    Arguments:
        timeout: Timeout to use when using ``select`` to block on the next input.
        constructor: The subtype of :class:`LogEntry` to build log entries with.
            By default it is just ``LogEntry``.
    """

    def __init__(self, timeout=0, constructor=None):
        self.processes = {}
        self.blocking = timeout is None
        self.timeout = timeout
        self.constructor = constructor or LogEntry
        self.buffer = []

    def start(self, argv, info, env=None, use_stdout=False, buffered=True, **options):
        """Start a process from the given ``argv``.

        Arguments:
            argv: The list of arguments.
            info: A dictionary of extra information that will be embedded in the
                ``LogEntry`` objects that are generated. The class given as the ``constructor``
                parameter should be able to take this information as keyword arguments
                in its ``__init__`` function.
            env: Environment variables to set, or None to pass ``os.environ``.
            use_stdout: If the program we are running is ``voir`` (``argv[0] == "voir"``)
                and ``use_stdout == True``, the ``DATA_FD`` environment variable will be
                set to 1, which is stdout. This will cause ``voir`` to smuggle data into
                the stdout file descriptor, and the Multiplexer will decode it
                transparently. See :mod:`voir.smuggle`.
            buffered: use to disable python output buffering.
                This is used to make test deterministic as buffering can cut lines are different spots.

        Returns:
            The subprocess object.
        """
        env = os.environ if env is None else env
        r, w = None, None
        buffered = "1" if buffered else "0"

        if use_stdout:
            proc = subprocess.Popen(
                argv,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env={**env, "DATA_FD": "1", "PYTHONUNBUFFERED": buffered},
                **options,
            )
            os.set_blocking(proc.stdout.fileno(), False)
            os.set_blocking(proc.stderr.fileno(), False)

            dec = Decoder(proc.stdout)
            mout = MultimodalFile(dec, "out", name=proc.stdout.name)
            mdat = MultimodalFile(dec, "data", name=proc.stdout.name)

            streams = [
                _Stream(pipe=mout, info={"pipe": "stdout"}, deserializer=None),
                _Stream(pipe=proc.stderr, info={"pipe": "stderr"}, deserializer=None),
                _Stream(pipe=mdat, info={"pipe": "data"}, deserializer=json.loads),
            ]

        else:
            r, w = os.pipe()
            proc = subprocess.Popen(
                argv,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                pass_fds=[w],
                env={**env, "DATA_FD": str(w), "PYTHONUNBUFFERED": buffered},
                **options,
            )
            readdata = open(r, "r", buffering=1)
            os.set_blocking(proc.stdout.fileno(), False)
            os.set_blocking(proc.stderr.fileno(), False)
            os.set_blocking(r, False)

            streams = [
                _Stream(pipe=proc.stdout, info={"pipe": "stdout"}, deserializer=None),
                _Stream(pipe=proc.stderr, info={"pipe": "stderr"}, deserializer=None),
                _Stream(pipe=readdata, info={"pipe": "data"}, deserializer=json.loads),
            ]

        self.add_process(
            proc=proc,
            argv=argv,
            info=info,
            streams=streams,
            w=w,
        )

        self.buffer.append(
            self.constructor(
                event="start",
                data={
                    "command": argv,
                    "time": time.time(),
                },
                **info,
            )
        )
        return proc

    def add_process(self, *, proc, info, argv, streams, w):
        """Add a process to those managed by this Multiplexer."""
        self.processes[proc] = (streams, argv, info, w)

    def _process_line(self, line, s, pinfo):
        try:
            if isinstance(line, bytes):
                line = line.decode("utf8")
            if s.deserializer:
                try:
                    data = s.deserializer(line)
                    if "$event" in data:
                        yield self.constructor(
                            event=data.pop("$event"),
                            data=data.pop("$data", None),
                            **data,
                            **pinfo,
                            **s.info,
                        )
                    else:
                        yield self.constructor(
                            event="data",
                            data=data,
                            **pinfo,
                            **s.info,
                        )
                except Exception as e:
                    yield self.constructor(
                        event="format_error",
                        data={
                            "line": line,
                            "type": type(e).__name__,
                            "message": str(e),
                        },
                        **pinfo,
                        **s.info,
                    )
            else:
                yield self.constructor(event="line", data=line, **pinfo, **s.info)
        except UnicodeDecodeError:
            yield self.constructor(event="binary", data=line, **pinfo, **s.info)

    def __iter__(self):
        """Iterate over all the events produced by the Multiplexer's processes."""
        yield from self.buffer
        self.buffer.clear()

        polling_obj = select.poll()
        to_consult = {}
        to_pipes = {}
        for proc, (streams, _, info, _) in self.processes.items():
            for s in streams:
                polling_obj.register(s.pipe, select.POLLIN | select.POLLPRI)
                entries = to_consult.setdefault(s.pipe, [])
                entries.append((s, proc, info))

                pipes = to_pipes.setdefault(s.pipe.fileno(), [])
                pipes.append(s.pipe)

        def close_resources(streams, w):
            for stream in streams:
                stream.pipe.close()

            if w is not None:
                try:
                    os.close(w)
                except Exception:
                    pass

        while self.processes:
            still_alive = set()
            ready = polling_obj.poll(self.timeout)

            for fd, event in ready:
                pipes = to_pipes[fd]
                for r in pipes:
                    while line := r.readline():
                        for s, proc, info in to_consult[r]:
                            yield from self._process_line(line, s, info)
                            still_alive.add(proc)

            for proc, (streams, argv, info, w) in list(self.processes.items()):
                if proc not in still_alive:
                    ret = proc.poll()
                    if ret is not None:
                        close_resources(streams, w)
                        del self.processes[proc]
                        yield self.constructor(
                            event="end",
                            data={
                                "command": argv,
                                "time": time.time(),
                                "return_code": ret,
                            },
                            **info,
                        )

            if not self.blocking:  # pragma: no cover
                yield None

        del polling_obj



import sys

from voir import give, iterate

if __name__ == "__main__":
    for x in iterate(
        "things",
        range(10, 15),
        report_batch=bool(int(sys.argv[1])),
        ignore_loading=True,
    ):
        give(x=x)



"""This module defines the Overseer, the main interface that instruments can use.

All instruments receive an :class:`Overseer` as their first argument.
"""

import importlib
import json
import os
import pkgutil
import sys
import traceback
from argparse import REMAINDER, Namespace
from pathlib import Path
from typing import Union

import yaml
from giving import Given, SourceProxy
from ptera import Probe, probing, select

from voir.smuggle import SmuggleWriter

from .argparse_ext import ExtendedArgumentParser
from .helpers import current_overseer
from .phase import GivenOverseer, Phase, PhaseSequence
from .scriptutils import resolve_script


class JsonlFileLogger:
    """Log data to a file as JSON lines.

    Arguments:
        filename: Either an integer representing a file descriptor to write to,
            or a path.
        require_writable: Require the file descriptor to be writable. If this is
            False and the file is not writable, this logger will simply forward
            the data to /dev/null instead of raising an OSError.
    """

    def __init__(self, filename, require_writable=True):
        self.filename = filename
        if self.filename == 1:
            self.out = SmuggleWriter(sys.stdout)
        elif self.filename == 2:
            self.out = SmuggleWriter(sys.stderr)
        else:
            try:
                self.out = open(self.filename, "w", buffering=1)
            except OSError:
                if require_writable:
                    raise
                self.out = open(os.devnull, "w")
        self.out.__enter__()

    def log(self, data):
        """Log a data dictionary as one JSON line into the file or file descriptor.

        If the data is not serializable as JSON, it will be dumped as
        ``{"$unserializable": repr(data)}``, and if _that_ fails, it will be
        dumped as the singularly uninformative ``{"$unrepresentable": None}``.
        """
        try:
            txt = json.dumps(data)
        except TypeError:
            try:
                txt = json.dumps({"$unserializable": repr(data)})
            except Exception:
                txt = json.dumps({"$unrepresentable": None})
        self.out.write(f"{txt}\n")

    def close(self):
        """Close the file."""
        self.out.__exit__()


class LogStream(SourceProxy):
    """Callable wrapper over :class:`giving.gvn.SourceProxy`.

    This has the same interface as :class:`giving.gvn.Given`.
    """

    def __call__(self, data):
        self._push(data)


class ProbeInstrument:
    """Instrument that creates a ptera.Probe on the given selector.

    The method ``overseer.probe()`` is shorthand for requiring an instance
    of this class.

    >>> probe = overseer.require(ProbeInstrument("f > x"))
    >>> probe.display()
    """

    def __init__(self, selector, **kwargs):
        self.selector = selector
        self.probe = self.__state__ = probing(self.selector, **kwargs)

    def __call__(self, ov):
        yield ov.phases.load_script(priority=0)
        with self.probe:
            yield ov.phases.run_script(priority=0)


class Overseer(GivenOverseer):
    """Oversee the running of a script and schedule instruments.

    When called with command-line arguments, the Overseer will parse instrument
    configuration, followed by the script to run (first positional argument), and
    then the script's arguments. Then it will load and run the script. Here is the
    sequence of phases. An instrument should yield a phase to wait until it is ended:

    * self.phases.init
        * Set up the logger and self.given
        * Parse the --config argument
    * self.phases.parse_args
        * Parse the command-line arguments
    * self.phases.load_script
        * Load the script's imports and functions
    * self.phases.run_script
        * Run the script
    """

    phases: PhaseSequence
    """Sequence of phases the Overseer goes through."""

    log: LogStream
    """A stream of data to log to $DATA_FD, possibly to the dashboard."""

    given: Given
    """A stream of data created by calls to :func:`~giving.api.give`."""

    argparser: ExtendedArgumentParser
    """The argument parser for voir, given before the script."""

    options: Namespace
    """The parsed arguments."""

    logfile: Union[str, int]
    """The name of the file to log to, or an integer file descriptor."""

    def __init__(self, instruments, logfile=None):
        """Initialize an Overseer.

        Arguments:
            instruments: Collection of instruments to require.
            logfile: Filename to log to, or integer file descriptor.
        """
        self.argparser = ExtendedArgumentParser()
        self.argparser.add_argument("SCRIPT", nargs="?", help="The script to run")
        self.argparser.add_argument(
            "ARGV", nargs=REMAINDER, help="Arguments to the script"
        )
        self.argparser.add_argument(
            "-m",
            dest="MODULE",
            nargs=REMAINDER,
            help="Module or module:function to run",
        )

        super().__init__(
            phase_names=["init", "parse_args", "load_script", "run_script", "finalize"],
            args=(self,),
            kwargs={},
        )
        self.require(*instruments)
        self.logfile = logfile

    def probe(self, selector: str, **kwargs) -> Probe:
        """Create a :class:`ProbeInstrument` on the given selector.

        >>> probe = overseer.probe("f > x")
        >>> probe.display()

        Arguments:
            selector: The selector to probe.
        """
        return self.require(ProbeInstrument(select(selector, skip_frames=1), **kwargs))

    def run_phase(self, phase: Phase):
        """Context manager to run a phase.

        >>> with self.run_phase(self.phases.peanuts):
        ...     do_stuff()

        The body of the with statement is run, and then all instruments that
        yielded that phase (in order to wait for its end) are resumed, in
        priority order.

        Arguments:
            phase: The phase to run.
        """
        self.log({"$event": "phase", "$data": {"name": phase.name}})
        return super().run_phase(phase)

    ####################
    # Internal methods #
    ####################

    def _on_instrument_error(self, e):
        self.log(
            {
                "$event": "overseer_error",
                "$data": {"type": type(e).__name__, "message": str(e)},
            }
        )
        print("=" * 80, file=sys.stderr)
        print(
            "voir: An error occurred in an overseer. Execution proceeds as normal.",
            file=sys.stderr,
        )
        print("=" * 80, file=sys.stderr)
        traceback.print_exception(type(e), e, e.__traceback__)
        print("=" * 80, file=sys.stderr)
        super()._on_instrument_error(e)

    def _run(self, argv):
        self.log = LogStream()
        self.given.where("$event") >> self.log
        if self.logfile is not None:
            self._logger = JsonlFileLogger(self.logfile, require_writable=False)
            self.log >> self._logger.log
        else:
            self._logger = None

        with self.run_phase(self.phases.init):
            tmp_argparser = ExtendedArgumentParser(add_help=False)
            tmp_argparser.add_argument("--config", action="append", default=[])
            tmp_options, argv = tmp_argparser.parse_known_args(argv)
            for config in tmp_options.config:
                self.argparser.merge_base_config(yaml.safe_load(open(config, "r")))

        with self.run_phase(self.phases.parse_args):
            self.options = self.argparser.parse_args(argv)
            del self.argparser

        with self.run_phase(self.phases.load_script):
            script, argv, func = _resolve_function(self.options)

        with self.run_phase(self.phases.run_script) as set_value:
            sys.argv = [script, *argv]
            set_value(func())

    def _prepare(self):
        super()._prepare()
        self._token = current_overseer.set(self)

    def _on_error(self, e):
        self.log(
            {
                "$event": "error",
                "$data": {"type": type(e).__name__, "message": str(e)},
            }
        )

    def _finish(self):
        super()._finish()
        with self.run_phase(self.phases.finalize):
            pass
        if self._logger:
            self._logger.close()
        current_overseer.reset(self._token)


def _resolve_function(options):
    """Resolve a function to call given an argparse options object.

    The relevant fields are ``(SCRIPT or MODULE) and ARGV``.
    """
    if script := options.SCRIPT:
        return script, options.ARGV, resolve_script(script)
    elif module_args := options.MODULE:
        module_spec, *argv = module_args
        if ":" in module_spec:
            module_name, field = module_spec.split(":", 1)
            module = importlib.import_module(module_name)
            return module_spec, argv, getattr(module, field)
        else:
            module_name = module_spec
            script = Path(pkgutil.get_loader(module_name).get_filename())
            if script.name == "__init__.py":
                script = script.parent / "__main__.py"
                module_name = f"{module_name}.__main__"
            script = str(script)
            return script, argv, resolve_script(script, module_name=module_name)
    else:
        sys.exit("Either SCRIPT or -m MODULE must be given.")



"""Utilities for instruments."""

import multiprocessing
import time
from threading import Thread


class Monitor(Thread):
    """Thread that calls a monitoring function every ``delay`` seconds."""

    def __init__(self, delay, func):
        super().__init__(daemon=True)
        self.stopped = False
        self.delay = delay
        self.func = func

    def run(self):
        while not self.stopped:
            time.sleep(self.delay)
            self.func()

    def stop(self):
        self.stopped = True


def _worker(state, queue, func, delay, init=None):
    if init is not None:
        init()

    while state["running"]:
        queue.put(func())
        time.sleep(delay)


class ProcessMonitor:
    """Thread monitor does not produce metrics on regular intervals,
    if the current process is used extensively, using the process monitor
    helps.
    """

    def __init__(self, delay, func, worker_init=None):
        self.manager = multiprocessing.Manager()
        self.state = self.manager.dict()
        self.state["running"] = True
        self.results = multiprocessing.Queue()
        self.process = multiprocessing.Process(
            target=_worker,
            args=(self.state, self.results, func, delay, worker_init),
        )

    def start(self):
        self.process.start()

    def stop(self):
        self.state["running"] = False
        self.process.join()


class ProcessPusher(Thread):
    def __init__(self, delay, queue, func):
        super().__init__(daemon=True)
        self.stopped = False

        self.delay = delay
        self.queue = queue
        self.func = func

    def push(self):
        while not self.queue.empty():
            self.func(self.queue.get())

    def run(self):
        while not self.stopped:
            time.sleep(self.delay)
            self.push()

    def stop(self):
        self.stopped = True
        self.push()


class _Monitor:
    def __init__(self, *args):
        self.args = args

    def start(self):
        for a in self.args:
            a.start()

    def stop(self):
        for a in self.args:
            a.stop()


def monitor(delay, getfun, pushfun, process=True, worker_init=None):
    """Run the monitor in a different process to have metrics in regular intervals
    Pusher is a thread that gets executed when there is time.

    delay: float
        sleep time between observation

    getfun: () -> dict
        function used to retrieve metrics about the system

    pushfun: (dict) -> None
        function used on the main process once matrics are received.
        It can be used to save/display the data

    process: bool
        If true it will instantiate a new process for the monitoring
    """
    if process:
        m = []

        # Note: monitor needs to be first, so it stops generating observation first
        monitor = ProcessMonitor(delay, getfun, worker_init)
        m.append(monitor)
        m.append(ProcessPusher(delay, monitor.results, pushfun))
        return _Monitor(*m)
    else:
        if worker_init:
            worker_init()

    def fun():
        pushfun(getfun())

    monitor = Monitor(delay, fun)
    return monitor



import threading

import pytest
from giving import give

from voir.phase import BaseOverseer, GivenOverseer, StopProgram


class LightOverseer(BaseOverseer):
    def __init__(self):
        self.results = []
        self.errors = []
        self.error_values = []
        super().__init__(
            ["one", "two", "three", "four"],
            args=[self, self.results],
            kwargs={},
        )

    def _on_instrument_error(self, err):
        if isinstance(err, AssertionError):
            self.stop(err)
        self.error_values.append(err)
        self.errors.append(type(err))

    def _run_phase(self, phase, value):
        with self.run_phase(phase) as set_value:
            if isinstance(value, Exception):
                self.results.append(type(value))
                raise value
            else:
                self.results.append(value)
                set_value(value)

    def _run(self, *values):
        self._run_phase(self.phases.one, values[0])
        self._run_phase(self.phases.two, values[1])
        self._run_phase(self.phases.three, values[2])
        self._run_phase(self.phases.four, values[3])
        self.results.append(5)


@pytest.fixture
def ov():
    return LightOverseer()


def test_single(ov):
    @ov.require
    def handler_appender(ov, seq):
        seq.append("zero")
        yield ov.phases.one
        seq.append("one")
        yield ov.phases.two
        seq.append("two")
        yield ov.phases.three
        seq.append("three")
        yield ov.phases.four
        seq.append("four")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == ["zero", 1, "one", 2, "two", 3, "three", 4, "four", 5]


def test_dual(ov):
    @ov.require
    def handler_A(ov, seq):
        seq.append("A0")
        yield ov.phases.one
        seq.append("A1")
        yield ov.phases.two
        seq.append("A2")
        yield ov.phases.three
        seq.append("A3")
        yield ov.phases.four
        seq.append("A4")

    @ov.require
    def handler_B(ov, seq):
        seq.append("B0")
        yield ov.phases.one
        seq.append("B1")
        yield ov.phases.two
        seq.append("B2")
        yield ov.phases.three
        seq.append("B3")
        yield ov.phases.four
        seq.append("B4")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [
        "A0",
        "B0",
        1,
        "A1",
        "B1",
        2,
        "A2",
        "B2",
        3,
        "A3",
        "B3",
        4,
        "A4",
        "B4",
        5,
    ]


def test_order(ov):
    @ov.require
    def handler_A(ov, seq):
        seq.append("A0")
        yield ov.phases.one
        seq.append("A1")
        yield ov.phases.two(priority=1)
        seq.append("A2")
        yield ov.phases.three(priority=-1)
        seq.append("A3")
        yield ov.phases.four
        seq.append("A4")

    @ov.require
    def handler_B(ov, seq):
        seq.append("B0")
        yield ov.phases.one
        seq.append("B1")
        yield ov.phases.two(priority=2)
        seq.append("B2")
        yield ov.phases.three
        seq.append("B3")
        yield ov.phases.four(priority=1)
        seq.append("B4")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [
        "A0",
        "B0",
        1,
        "A1",
        "B1",
        2,
        "B2",
        "A2",
        3,
        "B3",
        "A3",
        4,
        "B4",
        "A4",
        5,
    ]


def test_partial_phases(ov):
    @ov.require
    def handler_A(ov, seq):
        seq.append("A0")
        yield ov.phases.two
        seq.append("A2")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == ["A0", 1, 2, "A2", 3, 4, 5]


def test_add_multiple_copies(ov):
    def handler_A(ov, seq):
        seq.append("A0")
        yield ov.phases.two
        seq.append("A2")

    # Even though we add it 3 times, we should only execute handler_A once
    ov.require(handler_A)
    ov.require(handler_A)
    ov.require(handler_A)
    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == ["A0", 1, 2, "A2", 3, 4, 5]


def test_reenter(ov):
    @ov.require
    def handler_A(ov, seq):
        seq.append("A0")
        yield ov.phases.one
        seq.append("A1.1")
        yield ov.phases.one
        seq.append("A1.2")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == ["A0", 1, "A1.1", "A1.2", 2, 3, 4, 5]


def test_sandwiched_order(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.two(priority=10)
        seq.append("A2.1")
        yield ov.phases.two(priority=5)
        seq.append("A2.2")
        yield ov.phases.two(priority=-10)
        seq.append("A2.3")

    @ov.require
    def handler_B(ov, seq):
        yield ov.phases.two
        seq.append("B2.1")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [1, 2, "A2.1", "A2.2", "B2.1", "A2.3", 3, 4, 5]


def test_add_by_handler(ov):
    @ov.require
    def handler_A(ov, seq):
        seq.append("A0")
        yield ov.phases.one
        seq.append("A1")
        yield ov.phases.two(priority=1)
        seq.append("A2.1")
        hb = ov.require(handler_B)
        assert hb is handler_B
        seq.append("A2.2")
        yield ov.phases.three(priority=-1)
        seq.append("A3")
        yield ov.phases.four
        seq.append("A4")

    def handler_B(ov, seq):
        seq.append("B0")
        yield ov.phases.one
        seq.append("B1")
        yield ov.phases.two(priority=2)
        seq.append("B2")
        yield ov.phases.three
        seq.append("B3")
        yield ov.phases.four(priority=1)
        seq.append("B4")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [
        "A0",
        1,
        "A1",
        2,
        "A2.1",
        "B0",
        "B1",
        "A2.2",
        "B2",
        3,
        "B3",
        "A3",
        4,
        "B4",
        "A4",
        5,
    ]


def test_values(ov):
    @ov.require
    def handler_checker(ov, seq):
        one = yield ov.phases.one
        assert one == 1
        two = yield ov.phases.two
        assert two == 2
        three = yield ov.phases.three
        assert three == 3
        one_again = yield ov.phases.one
        assert one_again == 1
        four = yield ov.phases.four
        assert four == 4

    ov(1, 2, 3, 4)
    assert not ov.errors


def test_done(ov):
    @ov.require
    def handler_checker(ov, seq):
        yield ov.phases.one
        assert not ov.phases.one.done

        yield ov.phases.two
        assert ov.phases.one.done
        assert not ov.phases.two.done

        yield ov.phases.three
        assert ov.phases.two.done
        assert not ov.phases.three.done

        yield ov.phases.four
        assert ov.phases.three.done
        assert not ov.phases.four.done

    ov(1, 2, 3, 4)
    assert not ov.errors


def test_running(ov):
    @ov.require
    def handler_checker(ov, seq):
        assert not ov.phases.one.running
        yield ov.phases.one
        assert ov.phases.one.running

        assert not ov.phases.two.running
        yield ov.phases.two
        assert ov.phases.two.running
        assert not ov.phases.one.running

        assert not ov.phases.three.running
        yield ov.phases.three
        assert ov.phases.three.running
        assert not ov.phases.two.running

        assert not ov.phases.four.running
        yield ov.phases.four
        assert ov.phases.four.running
        assert not ov.phases.three.running

    ov(1, 2, 3, 4)
    assert not ov.errors


def test_runner_error(ov):
    @ov.require
    def handler_checker_1(ov, seq):
        try:
            yield ov.phases.two
        except TypeError:
            seq.append("error1")
            raise

    @ov.require
    def handler_checker_2(ov, seq):
        try:
            yield ov.phases.two
        except TypeError:
            seq.append("error2")
            raise RuntimeError("unrelated")

    @ov.require
    def handler_checker_3(ov, seq):
        try:
            yield ov.phases.two
        except TypeError:
            seq.append("error3")

    with pytest.raises(TypeError):
        ov(1, TypeError("uh oh"), 3, 4)

    assert ov.results == [1, TypeError, "error1", "error2", "error3"]
    assert ov.errors == [RuntimeError]


def test_handler_purposeful_error(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.two
        seq.append("A2")
        yield ov.phases.three
        seq.append("A3")

    @ov.require
    def handler_E(ov, seq):
        yield ov.phases.two
        ov.abort(RuntimeError("boom"))

    @ov.require
    def handler_B(ov, seq):
        yield ov.phases.two
        seq.append("B2")
        yield ov.phases.three
        seq.append("B3")

    with pytest.raises(RuntimeError):
        ov(1, 2, 3, 4)

    assert ov.results == [1, 2, "A2"]
    assert ov.errors == []


def test_immediate_handler_error(ov):
    @ov.require
    def handler_E(ov, seq):
        raise RuntimeError("boom")

    ov(1, 2, 3, 4)

    assert ov.results == [1, 2, 3, 4, 5]
    assert ov.errors == [RuntimeError]


def test_handler_not_a_generator(ov):
    @ov.require
    def handler_A(ov, seq):
        seq.append("A")

    ov(1, 2, 3, 4)

    assert not ov.errors
    assert ov.results == ["A", 1, 2, 3, 4, 5]


def test_bad_phase(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.one
        yield

    ov(1, 2, 3, 4)
    assert ov.errors == [Exception]

    with pytest.raises(Exception, match="must yield a valid phase"):
        raise ov.error_values[0]


def test_method(ov):
    class Handler:
        def __init__(self, letter):
            self.letter = letter

        def __call__(self, ov, seq):
            seq.append(f"{self.letter}0")
            yield ov.phases.one
            seq.append(f"{self.letter}1")
            yield ov.phases.two
            seq.append(f"{self.letter}2")
            yield ov.phases.three
            seq.append(f"{self.letter}3")
            yield ov.phases.four
            seq.append(f"{self.letter}4")

    ov.require(Handler("A"))
    ov.require(Handler("B"))
    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [
        "A0",
        "B0",
        1,
        "A1",
        "B1",
        2,
        "A2",
        "B2",
        3,
        "A3",
        "B3",
        4,
        "A4",
        "B4",
        5,
    ]


def test_state(ov):
    class Handler:
        def __init__(self, letter):
            self.__state__ = {"letter": letter}

        @property
        def letter(self):
            return self.__state__["letter"]

        def __call__(self, ov, seq):
            seq.append(f"{self.letter}0")
            yield ov.phases.one
            seq.append(f"{self.letter}1")
            yield ov.phases.two
            seq.append(f"{self.letter}2")
            yield ov.phases.three
            seq.append(f"{self.letter}3")
            yield ov.phases.four
            seq.append(f"{self.letter}4")

    @ov.require
    def handler_change(ov, seq):
        yield ov.phases.three
        h = ov.require(ha)
        assert h == {"letter": "A"}
        h["letter"] = "Z"

    ha = Handler("A")
    ov.require(ha)

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [
        "A0",
        1,
        "A1",
        2,
        "A2",
        3,
        "Z3",
        4,
        "Z4",
        5,
    ]


def test_immediate(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.IMMEDIATE
        seq.append("A")

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == ["A", 1, 2, 3, 4, 5]


def test_immediate_cannot_have_priority(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.IMMEDIATE(priority=1000)
        seq.append("A")

    ov(1, 2, 3, 4)
    assert ov.errors == [TypeError]


def test_stop(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.two
        ov.stop()

    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == [1, 2]
    assert ov.status == "stopped"


def test_stop_immediate(ov):
    def handler_A(ov, seq):
        ov.stop()

    ov.require(handler_A)
    ov(1, 2, 3, 4)
    assert not ov.errors
    assert ov.results == []
    assert ov.status == "stopped"


def test_stop_multiple_handlers(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.two
        ov.stop()

    @ov.require
    def handler_B(ov, seq):
        try:
            yield ov.phases.four
        except StopProgram:
            seq.append("B")

    @ov.require
    def handler_C(ov, seq):
        try:
            yield ov.phases.three
        except StopProgram:
            seq.append("C")
            raise RuntimeError("xxx")

    ov(1, 2, 3, 4)
    assert ov.errors == [RuntimeError]
    assert ov.results == [1, 2, "C", "B"]
    assert ov.status == "stopped"


def test_reenter_2(ov):
    @ov.require
    def handler_A(ov, seq):
        yield ov.phases.two
        ov(1, 2, 3, 4)  # error
        seq.append(111)

    ov(1, 2, 3, 4)
    assert ov.errors == [Exception]
    assert ov.results == [1, 2, 3, 4, 5]


def test_rerun(ov):
    ov(1, 2, 3, 4)
    with pytest.raises(Exception, match="Can only enter runner when"):
        ov(1, 2, 3, 4)


class LightGivenOverseer(GivenOverseer):
    def __init__(self):
        self.errors = []
        self.error_values = []
        super().__init__(
            ["one", "two", "three", "four"],
            args=[self],
            kwargs={},
        )

    def _run_phase(self, phase, value):
        with self.run_phase(phase) as set_value:
            if isinstance(value, Exception):
                self.give(error_type=type(value))
                raise value
            else:
                self.give(value=value)
                set_value(value)

    def _run(self, *values):
        self._run_phase(self.phases.one, values[0])
        self._run_phase(self.phases.two, values[1])
        self._run_phase(self.phases.three, values[2])
        self._run_phase(self.phases.four, values[3])
        self.give(value=5)


@pytest.fixture
def gov():
    return LightGivenOverseer()


def test_gov_give(gov):
    results = []

    @gov.require
    def handler_accumulator(ov):
        yield ov.phases.one
        ov.given["?value"].accum(results)

    @gov.require
    def handler_v(ov):
        yield ov.phases.two
        value = 13
        give(value)

    gov(1, 2, 3, 4)
    assert not gov.errors
    assert results == [2, 13, 3, 4, 5]


def test_gov_self_give(gov):
    results = []

    @gov.require
    def handler_accumulator(ov):
        yield ov.phases.one
        ov.given["?value"].accum(results)

    @gov.require
    def handler_v(ov):
        yield ov.phases.two
        value = 13
        ov.give(value=value)

    gov(1, 2, 3, 4)
    assert not gov.errors
    assert results == [2, 13, 3, 4, 5]


def test_gov_thread(gov):
    def q(ov):
        ov.give(value=4321)

    results = []

    @gov.require
    def handler_accumulator(ov):
        yield ov.phases.one
        ov.given["?value"].accum(results)

    @gov.require
    def handler_v(ov):
        thr = threading.Thread(target=q, args=(ov,))
        yield ov.phases.one
        thr.start()
        yield ov.phases.four
        thr.join()

    gov(1, 2, 3, 4)
    assert not gov.errors
    assert 4321 in results
